/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Define both _XOPEN_SOURCE and _GNU_SOURCE, this makes Solaris happier. Also
 * useful for systems without GNU_SOURCE. Need GNU_SOURCE for extra poll()
 * POLLRDHUP flag */
#define _XOPEN_SOURCE   500

#define _GNU_SOURCE

#ifdef __sun
/* For IPV6_V6ONLY we need either no POSIX XPG4_v2 or __EXTENSIONS__ */
# define __EXTENSIONS__
#endif

/* network functions */
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
/*#include <sys/select.h>*/
#include <poll.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <strings.h>

#include "net_common.h"
#include "xacml_io_ssl_log.h"

/************************************************************************
 * #define-s
 ************************************************************************/

#define HOSTNAME_MAX_CHARS      512
#define IP_MAX_CHARS            64
#define PORT_MAX_DIGITS         6
#if 0
#define PREFIX_V4_MAPPED        "::ffff:"
#endif


/************************************************************************
 * Static globals
 ************************************************************************/

static long xacml_socket_timeout      = 0;
static long xacml_connection_timeout  = 0;
static long xacml_incremental_backoff = 0;
static int  xacml_keepalive = 0;


/************************************************************************
 * Static prototypes
 ************************************************************************/

static int net_addr2host (char **                hostname,
                          socklen_t              bufsize,
                          const struct sockaddr *ss);

static int createAndSetUpATCPServerSocket(unsigned short port, int max_connections);


/************************************************************************
 * Function definitions
 ************************************************************************/

/* Sets TCP/IP socket() timeout in milli-seconds */
void xacml_io_set_socket_timeout(long input){
    xacml_socket_timeout = input;
}
/* Sets maximum time to spend on connecting to a host, including back-off retries */
void xacml_io_set_connection_timeout(long input){
    xacml_connection_timeout = input;
}
/* Sets initial incremental back-off time in milli-seconds */
void xacml_io_set_incremental_backoff(long input){
    xacml_incremental_backoff = input;
}
/* Sets initial TCP keepalive setting */
void xacml_io_set_keepalive(int input){
    xacml_keepalive = input;
}

/* Gets TCP/IP socket() timeout in milli-seconds */
long xacml_io_get_socket_timeout(void){
    return xacml_socket_timeout;
}
/* Gets maximum time to spend on connecting to a host, including back-off retries */
long xacml_io_get_connection_timeout(void){
    return xacml_connection_timeout;
}
/* Gets initial incremental back-off time in milli-seconds */
long xacml_io_get_incremental_backoff(void){
    return xacml_incremental_backoff;
}
/* Gets initial TCP keepalive setting */
int  xacml_io_get_keepalive(void){
    return xacml_keepalive;
}




/****************************************************
Function: xacml_io_setup_master_socket
Description:
          Setup Master Socket with all the socket options and such.

Variables:
          int * master_socket : master socket descriptor
          int   port          : port number to listen on
Return:
    0 : Good
   !0 : Not good
****************************************************/
int xacml_io_setup_master_socket (int * master_socket, int port)
{
    if ((port > 65535) || (port < 0))
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Failure in setting up listening socket: TCP port number out of range (0-65535) with value %d\n", __func__, port);
        return 1;
    }

    if ((*master_socket = createAndSetUpATCPServerSocket((unsigned short) port, 512)) < 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}


/****************************************************
Function: accept_new_client_socket
Description:
          Accepts new client socket from a master socket and returns that with the 
          DNS canonical name of the connected remote host

Variables:
          int master_socket   : master socket descriptor
          int * client_socket : return pointer to the client_socket descriptor
          char ** remote_host : DNS canonical name string of the remote host 
                                connecting to the master socket
Return:
    0 : Good
   !0 : Not good
****************************************************/
/* int manageClientConnectedOnServerSocket(int server_socket) */
int xacml_io_accept_new_client_socket (int                       server_socket, 
                                       int                     * client_socket, 
                                       char                   ** remote_host)
{
    struct sockaddr_storage sock_addr;
    /* struct sockaddr *client_socket_address = (struct sockaddr *)&sock_addr; */
    /* unsigned int client_socket_address_len = sizeof(sock_addr); */
    struct sockaddr *client_socket_address = (struct sockaddr *)&sock_addr;
    unsigned int client_socket_address_len = sizeof(sock_addr);

    /* accept connection */
    *client_socket = accept(server_socket, client_socket_address, &client_socket_address_len);

    if (*client_socket < 0)
    {
        XACML_LOG_FUNC(LOG_DEBUG, "%s: Client socket not accepted: %s\n", __func__, strerror(errno));
        return 1;
    }

    /* notify client connection */
    *remote_host = (char *) calloc (1, HOSTNAME_MAX_CHARS);
    net_addr2host(remote_host, HOSTNAME_MAX_CHARS, client_socket_address);

    return 0;
}



/**
 * Does the actual connect(). Reworked to use non-blocking I/O following example
 * in section 16.3 of Richard Stevens' Unix Network Programming.
 * Later further adapted to replace select() with poll() since select cannot
 * support (breaks) when socket fd > 1024
 */
int xacml_io_create_client_socket (int * client_socket,
				   const char * server,
				   int port,
				   long time_out_milliseconds)
{
    struct addrinfo  hints;
    struct addrinfo *res;
    int              rc;
    int              mysock = -1;
    char             portstr[24];
    int		     flags;
    struct pollfd    fds;
    socklen_t	     len;
    int		     sock_err=0;
    /* split-up time-out into msec and seconds */
    long             msecs = time_out_milliseconds % 1000; /* msec < 1000 */
    long             secs  = (time_out_milliseconds-msecs)/1000; /* seconds */
    long             usecs = msecs*1000;    /* microseconds */
    struct timeval   wait_tv;
    int              optval = 0;
    socklen_t        optlen;

    
    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_socktype = SOCK_STREAM;
    hints.ai_family   = AF_INET;

    /* Get addrinfo */
    snprintf(portstr, (size_t)24, "%d", port);
    rc = getaddrinfo(server, &portstr[0], &hints, &res);
    if (rc != 0) {
        XACML_LOG_FUNC(LOG_ERR,
		"%s: Error: Failed to getaddrinfo (%s, %s, *, *): %s\n",
		__func__,server, portstr,
		(rc==EAI_SYSTEM ? strerror(errno): gai_strerror(rc)) );
        return 2;
    }

    /* Create new socket */
    mysock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (mysock < 0) {
        XACML_LOG_FUNC(LOG_ERR, "%s: Error: Failed to create socket: %s\n",
		__func__,strerror(errno));
        freeaddrinfo( res );
        return 3;
    }

    /* Set to non-blocking I/O, use select() to check whether connect is
     * finished */
    flags=fcntl(mysock, F_GETFL, 0);
    fcntl(mysock, F_SETFL, flags | O_NONBLOCK);

    /* non-blockingly connecting socket to host on port, timeout is handled at
     * select(). Note that on Linux a connect() timeout would listen to a
     * SO_SNDTIMEO setting using setsockopt(). On Solaris or Mac this doesn't
     * work. */
    rc=connect(mysock, res -> ai_addr, res -> ai_addrlen);
    if (rc<0)	{
	if (errno!=EINPROGRESS)	{
	    XACML_LOG_FUNC(LOG_ERR, "%s: connect() failed: %s\n",
		    __func__, strerror(errno));
	    goto connect_failure;
	}

	/* set requested flags for poll() */
	fds.fd=mysock;
#ifdef POLLRDHUP
	fds.events = POLLIN | POLLOUT | POLLRDHUP;
#else
	fds.events = POLLIN | POLLOUT;
#endif
	fds.revents=0; /* Not clear if this is needed */

	XACML_LOG_FUNC(LOG_DEBUG,
		    "%s: Calling poll() with timeout: %ld msec\n",
		    __func__, time_out_milliseconds);

	/* Call poll() on mysock with given timeout */
	rc=poll(&fds, (nfds_t)1, (int)time_out_milliseconds);
	if (rc<=0)  {	/* error or timeout */
	    if (rc<0)	/* error */
		XACML_LOG_FUNC(LOG_ERR, "%s: poll() failed: %s\n", __func__,
		    strerror(errno));
	    else    /* timeout */
		XACML_LOG_FUNC(LOG_ERR, "%s: poll() timed-out\n", __func__);
	    close(mysock);
	    goto connect_failure;
	}

	/* First check POLLNVAL error: invalid fd, so shouldn't close socket */
	if (fds.revents & POLLNVAL) {
	    XACML_LOG_FUNC(LOG_ERR, 
		    "%s: poll(): invalid request: fd not open (output only)\n",
		    __func__);
	    goto connect_failure;
	}

	/* Now check for any pending errors. Note that some platforms return 0,
	 * some -1 for error, so check both rc and sock_err */
	len = sizeof(sock_err);
	if (getsockopt(mysock, SOL_SOCKET, SO_ERROR, &sock_err, &len)<0 ||
	    sock_err)
	{
	    if (sock_err)
		XACML_LOG_FUNC(LOG_ERR, "%s: socket error: %s\n",
			__func__, strerror(sock_err));
	    else
		XACML_LOG_FUNC(LOG_ERR, "%s: getsockopt() failed: %s\n",
			__func__, strerror(errno));
	    close(mysock);
	    goto connect_failure;
	}

	/* Check POLL(RD)HUP errors: remote hang up. This isn't always caught
	 * by getsockopt(). Check first for getsockopt() since it's more
	 * instructive. Linux: disconnect after connect(): POLLRDHUP /
	 * getsockopt() without errors. disconnect before connect(): POLLHUP /
	 * getsockopt() with error. */
#ifdef POLLRDHUP
	if ( (fds.revents & POLLHUP)!=0 || (fds.revents & POLLRDHUP)!=0 )
#else
	if ( (fds.revents & POLLHUP)!=0 )
#endif
	{
	    XACML_LOG_FUNC(LOG_ERR, "%s: poll(): peer closed connection.\n",
		    __func__);
	    close(mysock);
	    goto connect_failure;
	}

	/* This is probably not needed (getsockopt probably has caught it), but
	 * check anyway */
	if ((fds.revents & POLLERR)!=0) {
	    XACML_LOG_FUNC(LOG_ERR,
		    "%s: poll(): Error condition (output only)\n", __func__);
	    close(mysock);
	    goto connect_failure;
	}
    }

    /* restore file status flags */
    fcntl(mysock, F_SETFL, flags);

    /* Setup read/write timeout for socket */
    wait_tv.tv_sec =  secs;
    wait_tv.tv_usec = usecs;
    optlen = sizeof(wait_tv);
    XACML_LOG_FUNC(LOG_DEBUG,
	    "%s: Calling setsockopt() with timeout: %ld sec and %ld usec\n",
	    __func__,(long)(wait_tv.tv_sec),(long)(wait_tv.tv_usec));
    if (setsockopt (mysock, SOL_SOCKET, SO_RCVTIMEO, &wait_tv, optlen) < 0 ||
	setsockopt (mysock, SOL_SOCKET, SO_SNDTIMEO, &wait_tv, optlen) < 0)
    {
        XACML_LOG_FUNC(LOG_ERR,
		"%s: Error: Failed to set the socket timeout: %s\n",
		__func__,strerror(errno));
	goto connect_failure;
    }

    /* Socket is succesfully connected - configing TCP keep-alive */
    switch (xacml_io_get_keepalive()) {
	case XACML_KEEPALIVE_ENABLED :
	    optval = 1;
	    optlen = sizeof(optval);
	    setsockopt (mysock, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen);
	    break;
	case XACML_KEEPALIVE_DISABLED :
	    optval = 0;
	    optlen = sizeof(optval);
	    setsockopt (mysock, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen);
	    break;
	case XACML_KEEPALIVE_SYSTEMDEFAULT :
	    /* Do nothing, because we now rely on the system default for
	     * instance */
	    break;
    }

    *client_socket = mysock;
    freeaddrinfo( res );
    return 0;

connect_failure:
    freeaddrinfo( res );
    return 6;
}

/************************************************************************
 * Static functions
 ************************************************************************/

/***************************************************
Function:
    createAndSetUpATCPServerSocket
Description:
    This function is a direct descendant of the examples by the 
    EGEE-III SA2 group.
    The function will return with -1 or with an active bound and listening master socket
    which will listen for IPv4 and IPv6 clients to connect.
 
Parameters:
    port                Sets the TCP portnumber on which the service will listen
                        The rang is from 0 to 65535
    max_connections     Sets the amount of listener sockets for the process to preload
                        and which can be dispatched to accept new connections.
***************************************************/
static int createAndSetUpATCPServerSocket(unsigned short port, int max_connections)
{

    int server_socket=-1;
    int success=0;
    struct addrinfo hints, *address_list = NULL, *paddress;
    int on = 1, off = 0;
    char * ip = NULL;
    char port_string[PORT_MAX_DIGITS];
    int rc;

    /* getaddrinfo parameters */
    memset(&hints, 0, sizeof(hints));   /* init */
    hints.ai_flags |= AI_PASSIVE;       /* server behaviour: accept connection
					   on any network address */
    hints.ai_family = AF_UNSPEC;        /* find both ipv4 and ipv6 addresses */
    hints.ai_socktype = SOCK_STREAM;    /* type of socket */

    rc=snprintf(port_string, PORT_MAX_DIGITS, "%d", port);
    if (rc<0)   {
	XACML_LOG_FUNC(LOG_ERR, "%s: Error converting port %d to string\n",
		__func__, port);
	return -1;
    }
    if (rc>=PORT_MAX_DIGITS)    {
	XACML_LOG_FUNC(LOG_ERR, "%s: port %d does not fit in char[%d]\n",
		__func__ ,port, PORT_MAX_DIGITS);
	return -1;
    }

    rc = getaddrinfo(NULL, port_string, &hints, &address_list);
    if (rc != 0)
    {
        XACML_LOG_FUNC(LOG_ERR,
		"%s: Error: Failed to getaddrinfo(NULL,%s, *, *): %s\n",
		__func__,port_string,
		(rc==EAI_SYSTEM ? strerror(errno): gai_strerror(rc)) );
	return -1;
    }

    for (paddress = address_list; paddress; paddress = paddress->ai_next)   {
	if (server_socket!=-1)	{
	    close(server_socket);
	    server_socket = -1;
	}

        /* create the socket */
        server_socket = socket(paddress->ai_family,
			       paddress->ai_socktype,
			       paddress->ai_protocol);
        if (server_socket == -1)    {
	    XACML_LOG_FUNC(LOG_WARNING,"%s: Cannot create socket: %s\n",
		    __func__, strerror(errno));
            continue; /* try next address */
        }

        /* set SO_REUSEADDR option */
        if (setsockopt(server_socket,
		       SOL_SOCKET, SO_REUSEADDR, &on, (socklen_t)sizeof(int))==-1) {
	    XACML_LOG_FUNC(LOG_WARNING,"%s: Cannot set SO_REUSEADDR: %s\n",
		    __func__, strerror(errno));
	    continue;
	}

        /* we must unset IPV6_V6ONLY in order to allow IPv4 clients
	 * to connect to the IPv6 socket */
        if (paddress->ai_family == AF_INET6)
            if (setsockopt(server_socket,
			   IPPROTO_IPV6, IPV6_V6ONLY, &off, (socklen_t)sizeof(int))==-1) {
		XACML_LOG_FUNC(LOG_WARNING,"%s: Cannot unset IPV6_V6ONLY: %s\n",
			__func__, strerror(errno));
		continue;
	    }

        /* get the ip string, display it with the port number  */
        net_addr2host (&ip, IP_MAX_CHARS, paddress->ai_addr);

        /* bind and listen */
        if ( (bind(server_socket,
		   paddress->ai_addr, paddress->ai_addrlen) == -1) ||
             (listen(server_socket, max_connections) == -1)) {
	    XACML_LOG_FUNC(LOG_NOTICE,
		    "%s: Cannot bind on %s port %d: %s\n",
		    __func__, ip, port, strerror(errno));
            continue; /* try next address */
        }
        
	XACML_LOG_FUNC(LOG_INFO,
		"%s: Server socket now listening on %s port %d\n",
		__func__, ip, port);

	/* Make sure to set success flag */
	success=1;

	/* break if no error */
        break;
    }

    /* Close socket if unsuccesful */
    if (!success && server_socket!=-1)	{
	close(server_socket);
	server_socket = -1;
    }

    /* free ip */
    free(ip);

    /* free the list */
    freeaddrinfo(address_list);

    /* return the server socket */
    return server_socket;
}


/****************************************************
Function: net_addr2host
Description:
          Will convert a sockaddr_storage struct to
          a hostname, through reverse Host and DNS lookup
          or it will return the IP address as string.

          This function originated from the EGEE-III SA2 team as i
          IPv6 example and is altered to work of its new context
Variables:
            char ** hostname            : buffer in which the hostname will be written,
                                          Will allocate if not allocated yet.
            socklen_t bufsize           : size of hostname buffer
            struct  sockaddr_storage ss : socket address information (for Internet)
Return:
            0 : Good
            !0 : Not good
****************************************************/
static int net_addr2host (char **                hostname,
                          socklen_t              bufsize,
                          const struct sockaddr *ss)
{
    int rc = 0;

    /* Allocate if you didn't yet */
    if (*hostname == NULL)
        *hostname = (char *) calloc (1,sizeof (char) * bufsize);

    /* Try to get a hostname resolution from the IPv6 compliant sockaddr container.
     * Note 1: when not specifying NI_NAMEREQD as flag, it automatically
     * converts to normal IP representation if no hostname can be found */
    /* Note 2: we cannot easily get rid of the implicit cast from size_t to
     * socklen_t (which is not a long unsigned). Sizeof a struct
     * sockaddr_storage better fits in a socklen_t, so in practise we should be
     * fine */
    /* Note 3: bufsize is described in many Linux manpages as a size_t but is
     * actually nowadays a socklen_t */ 
    if ((rc = getnameinfo (ss,
			   (socklen_t)sizeof (struct sockaddr_storage),
                           *hostname, bufsize,
                           NULL, 0, 0) != 0))	{
	XACML_LOG_FUNC(LOG_DEBUG, "%s: getnameinfo() failed: %s\n",
		__func__,
		(rc==EAI_SYSTEM ? strerror(errno): gai_strerror(rc)) );
	return 1;
    }
    return 0;
}
